# Copyright 2023, Méline Sieber
#
# This file is part of Nobubo.
#
# Nobubo is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Nobubo is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with Nobubo.  If not, see <https://www.gnu.org/licenses/>.

"""
Contains functions for various output layouts.
"""
import logging
import pathlib
import random
import string
import subprocess
from dataclasses import dataclass
from typing import List, Tuple

from nobubo import errors


logger = logging.getLogger(__name__)


@dataclass
class PageSize:
    """
    Page size whose width and height are in user space units.
    """

    width: float = 0
    height: float = 0


@dataclass
class Layout:
    """
    A Pattern layout.

    first_page: The number of the pdf page
    which marks the beginning of the pattern pages
    that are covered by the columns and rows.
    """

    first_page: int
    columns: int
    rows: int


class NobuboInput:
    """
    Holds all information of the input pdf.
    """

    def __init__(
        self,
        input_filepath: pathlib.Path,
        number_of_pages: int,
        pagesize: PageSize,
        layout: List[Layout],
        reverse_assembly: bool = False,
    ):
        """
        Holds all information concerning the input pdf and is responsible
        for creating the collage.
        :param input_filepath: path to the input pdf
        :param number_of_pages: total amount of pages in the pdf
        :param pagesize: width and height of a pdf page in user space units
        :param layout: layout of the pdfs
        :param reverse_assembly: False: assemble pdf from top left to bottom right,
        True: assemble pdf from bottom left to top right.
        """
        self.input_filepath = input_filepath
        self.number_of_pages = number_of_pages
        self.pagesize = pagesize
        self.layout = layout
        self.reverse_assembly = reverse_assembly

    def __repr__(self):
        return (
            f"<class '{self.__class__.__name__}': "
            f"input_filepath: '{self.input_filepath}', "
            f"number_of_pages: '{self.number_of_pages}', "
            f"pagesize: '{self.pagesize}', "
            f"layout: '{self.layout}', "
            f"reverse_assembly: '{self.reverse_assembly}'>"
        )

    def assemble_collage(self, temp_output_dir: pathlib.Path) -> List[pathlib.Path]:
        """
        Takes a pattern pdf where one page equals a part of the pattern and
        assembles it to one huge collage.
        The default assembles it from top left to the bottom right.
        :param temp_output_dir: The temporary path where all calculations should happen.
        :return A list of all the path to the collages, each with all pattern pages
                assembled on one single page.

        """
        all_collages_paths: List[pathlib.Path] = []
        for counter, current_layout in enumerate(self.layout):
            logging.info(f"Assembling overview {counter + 1} of {len(self.layout)}\n")
            logging.info("Creating collage...")
            all_collages_paths.append(self._assemble(temp_output_dir, current_layout))
        return all_collages_paths

    def _assemble(
        self, temp_output_dir: pathlib.Path, current_layout: Layout
    ) -> pathlib.Path:
        collage_width = self.pagesize.width * current_layout.columns
        collage_height = self.pagesize.height * current_layout.rows

        if self.reverse_assembly:
            logging.debug("Reverse assembly chosen")
            start, end, step = reverse_pagerange(current_layout)
            page_range_for_pdflatex = list(
                reversed(
                    [
                        (x, x + current_layout.columns - 1)
                        for x in range(start, end, step)
                    ]
                )
            )
            tuples = ["-".join(map(str, i)) for i in page_range_for_pdflatex]
            page_range = ",".join(tuples)
        else:
            begin = current_layout.first_page
            end_of_section = current_layout.columns * current_layout.rows
            end = current_layout.first_page + end_of_section - 1
            page_range = f"{begin}-{end}"

        file_content = [
            "\\batchmode\n",
            "\\documentclass[a4paper,]{article}\n",
            f"\\usepackage[papersize={{{collage_width}pt,"
            f"{collage_height}pt}}]{{geometry}}\n",
            "\\usepackage[utf8]{inputenc}\n",
            "\\usepackage{pdfpages}\n",
            "\\begin{document}\n",
            f"\\includepdfmerge[nup={current_layout.columns}x{current_layout.rows}, "
            f"noautoscale=true, scale=1.0]"
            f"{{{str(self.input_filepath)},{page_range} }}\n",
            "\\end{document}\n",
        ]

        input_filepath = temp_output_dir / "texfile.tex"
        output_filename = f"output_{random_string()}"

        with input_filepath.open("w") as f:
            f.writelines(file_content)

        command = [
            "pdflatex",
            "-interaction=nonstopmode",
            f"-jobname={output_filename}",
            f"-output-directory={temp_output_dir}",
            str(input_filepath),
        ]

        logging.debug("Sending command to pdflatex")
        try:
            subprocess.check_output(command, stderr=subprocess.STDOUT)
        except subprocess.CalledProcessError as e:
            raise errors.UsageError(
                "Error: pdflatex encountered a problem while "
                f"assembling the collage and had to abort:\n{e}"
            )
        except FileNotFoundError as e:
            raise errors.UsageError(f"pdflatex or the output file was not found:\n{e}")

        return temp_output_dir / pathlib.Path(output_filename).with_suffix(".pdf")


def reverse_pagerange(layout: Layout) -> Tuple[int, int, int]:
    return (
        layout.first_page,
        layout.first_page + (layout.columns * layout.rows) - 1,
        layout.columns,
    )


def random_string():
    return "".join(random.choices(string.ascii_lowercase + string.digits, k=7))



# Copyright 2023, Méline Sieber
#
# This file is part of Nobubo.
#
# Nobubo is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Nobubo is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with Nobubo.  If not, see <https://www.gnu.org/licenses/>.
import logging
import pathlib
import re
import sys
import tempfile
from typing import List

import click

from nobubo import errors
from nobubo.init_nobubo import parse_cli_input_data, parse_cli_output_data


logger = logging.getLogger(__name__)


def validate_output_layout(ctx, param, value):
    p = re.compile(r"(a0)|(us)|(\d+[x]\d+)")
    try:
        assert value is None or p.match(value)
        return value
    except AssertionError:
        raise click.BadParameter(
            f"Output layout {value} does not exist. "
            "Have you chosen a0, us or a custom layout, "
            "such as 222x444?"
        )


@click.command()
@click.option(
    "--il",
    "input_layout_cli",
    nargs=3,
    type=click.INT,
    multiple=True,
    required=True,
    help="Input layout of the pdf. Can be used multiple times.",
    metavar="FIRSTPAGE COLUMNS ROWS",
)
@click.option(
    "--ol",
    "output_layout_cli",
    nargs=1,
    type=click.STRING,
    callback=validate_output_layout,
    help="Output layout. Supported formats: a0, us, custom. No output "
    "layout provided creates a huge collage.",
    metavar="a0 | us | mmxmm",
)
@click.option(
    "--margin",
    "print_margin",
    nargs=1,
    type=click.INT,
    help="Define an optional print margin in mm.",
    metavar="mm",
)
@click.option(
    "--reverse",
    "reverse_assembly",
    is_flag="True",
    help="With reverse flag: collage is assembled from bottom left to top right. "
    "No flag: collage is assembled from top left to bottom right. ",
)
@click.argument("input_path", type=click.STRING)
@click.argument("output_path", type=click.STRING)
def main(
    input_layout_cli,
    output_layout_cli,
    print_margin,
    reverse_assembly,
    input_path,
    output_path,
):
    """
    Creates a collage from digital pattern pages
    and then chops it up into a desired output layout.
    The collage is assembled according to one or several overview sheets.
    These overviews are usually provided along with the pattern pages
    in the same pdf or in the instructions pdf.

    Note: In order to use nobubo, you need the original pdf pattern.
    Create a backup of the original if you are afraid to have it damaged in any way.
    The author takes no responsibility if you face any fit issues
    or other problems now or later on.

    Example: A digital pattern contains 2 overview sheets
    at page 1 and 34 with different layouts each.
    The output is to be printed on A0 paper:

    $ nobubo --il 2 8 4 -il 35 7 3 --ol a0 "myfolder/mypattern.pdf" "test_collage.pdf"

    See the readme for further information: https://github.com/bytinbit/nobubo

    Arguments:

    INPUT_PATH: Input path to your pdf pattern.

    OUTPUT_PATH: Where the output should be saved.

    """
    logging.basicConfig(
        level=logging.INFO,
        format="%(message)s",
    )
    try:
        nobubo_input = parse_cli_input_data(
            input_layout_cli, reverse_assembly, input_path
        )
        nobubo_output = parse_cli_output_data(
            output_layout_cli, print_margin, output_path
        )
        with tempfile.TemporaryDirectory() as td:
            temp_output_dir = pathlib.Path(td)
            temp_collage_paths: List[pathlib.Path] = nobubo_input.assemble_collage(
                temp_output_dir
            )
            logger.info(f"Successfully assembled collage from {input_path}.\n")
            if nobubo_output.output_pagesize is not None:
                nobubo_output.create_output_files(temp_collage_paths, nobubo_input)
            else:  # default: no output_layout specified, print collage pdf
                nobubo_output.write_collage(
                    temp_collage_paths,
                )
            print("All done, enjoy your sewing! :)")

    except (errors.UsageError, click.BadParameter) as e:
        print(e)
        sys.exit(1)



# Copyright 2023, Méline Sieber
#
# This file is part of Nobubo.
#
# Nobubo is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Nobubo is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with Nobubo.  If not, see <https://www.gnu.org/licenses/>.

"""
Contains functions for various output layouts.
"""
import logging
import math
import pathlib
from copy import copy
from dataclasses import dataclass
from typing import List, Tuple, Optional

import pikepdf

from nobubo import errors
from nobubo import assembly

logger = logging.getLogger(__name__)


@dataclass
class Point:
    """
    Point on a pdf page in user space units.
    """

    x: float
    y: float


@dataclass
class Factor:
    """
    Factor class for multiplication.
    """

    x: int
    y: int


class NobuboOutput:
    """
    Holds all information of the output pdf and is responsible for creating
    the desired output pdf.
    """

    def __init__(
        self, output_path: pathlib.Path, output_pagesize: Optional[assembly.PageSize]
    ):
        """
        :param output_path: path where the output pdf should be saved.
        :param output_pagesize: The desired page size in user space units (can include
        user-defined print margin).
        """
        self.output_path = output_path
        self.output_pagesize = output_pagesize

    def __repr__(self):
        return (
            f"<class '{self.__class__.__name__}': "
            f"output_path: '{self.output_path}', "
            f"output_pagesize: '{self.output_pagesize}'>"
        )

    def create_output_files(
        self,
        temp_collage_paths: List[pathlib.Path],
        input_properties: assembly.NobuboInput,
    ) -> None:
        for counter, collage_path in enumerate(temp_collage_paths):
            try:
                collage = pikepdf.Pdf.open(collage_path)
            except OSError as e:
                raise errors.UsageError(
                    "Could not open collage file for disassembly:" f"\n{e}."
                )
            new_outputpath = self.generate_new_outputpath(self.output_path, counter)
            logger.debug("Chopping up collage")
            chopped_up_files = self._create_output_files(
                collage,
                input_properties.pagesize,
                input_properties.layout[counter],
            )
            logger.debug("Successfully chopped up the collage.\n")
            self.write_chops(chopped_up_files, new_outputpath)
            logger.info(f"Final pdf written to {new_outputpath}.\n")

    def write_chops(self, collage: pikepdf.Pdf, output_path: pathlib.Path) -> None:
        logger.info("Writing files...")
        try:
            collage.save(output_path)
        except OSError as e:
            raise errors.UsageError(
                f"An error occurred " f"while writing the output file:\n{e}"
            )

    def write_collage(
        self,
        temp_collage_paths: List[pathlib.Path],
    ) -> None:
        for counter, collage_path in enumerate(temp_collage_paths):
            new_outputpath = self.generate_new_outputpath(self.output_path, counter)
            try:
                temp_collage = pikepdf.Pdf.open(collage_path)
                temp_collage.save(new_outputpath)
            except OSError as e:
                raise errors.UsageError(
                    f"An error occurred " f"while writing the collage:\n{e}"
                )
            logger.info(f"Collage written to {new_outputpath}.")

    def _create_output_files(
        self,
        collage: pikepdf.Pdf,
        input_pagesize: assembly.PageSize,
        current_layout: assembly.Layout,
    ) -> pikepdf.Pdf:
        """
        Chops up the collage that consists of all the pattern pages to individual pages
        of the desired output size.
        :param collage: One pdf page that contains all assembled pattern pages.
        :param input_pagesize: size of an input pdf page
        :param current_layout: the current layout of the input pdf
        :return: The pdf with several pages, ready to write to disk.
        """
        logger.info("Using collage to create desired output layout")
        assert self.output_pagesize is not None
        n_up_factor = self.nup_factors(input_pagesize, self.output_pagesize)
        # only two points are needed to be cropped,
        # lower left (x, y) and upper right (x, y)
        lowerleft_factor = Factor(x=0, y=0)
        upperright_factor = Factor(x=1, y=1)

        output = pikepdf.new()
        # pdfstitcher made me aware of pikepdf and provided some hints
        # on how to use it, thanks!
        # https://github.com/cfcurtis/pdfstitcher

        for i in range(0, self.pages_needed(current_layout, n_up_factor)):
            page = copy(collage.pages[0])

            lowerleft: Point = _calculate_lowerleft_point(
                lowerleft_factor, n_up_factor, input_pagesize
            )
            upperright: Point = _calculate_upperright_point(
                upperright_factor, n_up_factor, current_layout, input_pagesize
            )

            # adjust multiplying factor
            colsleft = _calculate_colsrows_left(
                current_layout.columns, upperright_factor.x, n_up_factor.x
            )
            lowerleft_factor, upperright_factor = _adjust_factors(
                lowerleft_factor, upperright_factor, colsleft
            )

            page.CropBox = [lowerleft.x, lowerleft.y, upperright.x, upperright.y]
            output.pages.append(page)

        return output

    def pages_needed(self, layout: assembly.Layout, n_up_factor: Factor) -> int:
        """
        Calculate the pages needed for the required output layout.
        :param layout: layout of the input pdf
        :param n_up_factor: how many pages of the input pdf fit on the desired layout
        :return:
        """
        x = layout.columns / n_up_factor.x
        y = layout.rows / n_up_factor.y
        return math.ceil(x) * math.ceil(y)

    def nup_factors(
        self, input_pagesize: assembly.PageSize, output_pagesize: assembly.PageSize
    ) -> Factor:
        """
        Calculate the n-up factor for the output pdf, i.e. how many input pages
        fit on the desired output layout.
        :param input_pagesize: Size of a page of the input pdf in user space units
        :param output_pagesize: the output layout in user space units
        :return:
        """
        x_factor = int(output_pagesize.width // input_pagesize.width)
        y_factor = int(output_pagesize.height // input_pagesize.height)
        return Factor(x=x_factor, y=y_factor)

    def generate_new_outputpath(
        self, output_path: pathlib.Path, page_count: int
    ) -> pathlib.Path:
        new_filename = f"{output_path.stem}_{page_count + 1}{output_path.suffix}"
        return output_path.parent / new_filename


def _calculate_colsrows_left(layout_element: int, factor: int, nup_factor: int) -> int:
    return layout_element - (factor * nup_factor)


def _calculate_lowerleft_point(
    lowerleft_factor: Factor, n_up_factor: Factor, pagesize: assembly.PageSize
) -> Point:
    return Point(
        x=lowerleft_factor.x * n_up_factor.x * pagesize.width,
        y=lowerleft_factor.y * n_up_factor.y * pagesize.height,
    )


def _calculate_upperright_point(
    upperright_factor: Factor,
    n_up_factor: Factor,
    current_layout: assembly.Layout,
    pagesize: assembly.PageSize,
) -> Point:
    upperright = Point(x=0, y=0)
    # Manage ROWS: apply transformation to upper right, y-value
    rowsleft = _calculate_colsrows_left(
        current_layout.rows, upperright_factor.y, n_up_factor.y
    )
    if rowsleft < 0:  # end of pattern reached  (full amount of rows reached)
        upperright.y = current_layout.rows * pagesize.height
    else:
        upperright.y = upperright_factor.y * n_up_factor.y * pagesize.height

    # Manage COLS: apply transformation to upper right, x-value
    colsleft = _calculate_colsrows_left(
        current_layout.columns, upperright_factor.x, n_up_factor.x
    )  # COLS
    if colsleft > 0:  # still assembling the same horizontal line
        upperright.x = upperright_factor.x * n_up_factor.x * pagesize.width

    else:  # end of line reached, need to go 1 row up
        if colsleft == 0:  # cols % n_up_factor == 0
            upperright.x = upperright_factor.x * n_up_factor.x * pagesize.width
        if colsleft < 0:  # remainder pages left for COLS
            upperright.x = current_layout.columns * pagesize.width
    return upperright


def _adjust_factors(
    lowerleft_factor: Factor, upperright_factor: Factor, colsleft: int
) -> Tuple[Factor, Factor]:
    if colsleft > 0:  # still assembling the same horizontal line
        return _advance_horizontally(lowerleft_factor, upperright_factor)
    else:  # end of line reached, need to go 1 row up
        return _advance_vertically(lowerleft_factor, upperright_factor)


def _advance_horizontally(
    lowerleft_factor: Factor, upperright_factor: Factor
) -> Tuple[Factor, Factor]:
    lowerleft_factor.x += 1
    upperright_factor.x += 1
    return lowerleft_factor, upperright_factor


def _advance_vertically(
    lowerleft_factor: Factor, upperright_factor: Factor
) -> Tuple[Factor, Factor]:
    lowerleft_factor.x = 0
    lowerleft_factor.y += 1

    upperright_factor.x = 1
    upperright_factor.y += 1
    return lowerleft_factor, upperright_factor



"""
Entry point module for the execution of `python -m nobubo`
See:
https://docs.python.org/3/using/cmdline.html#cmdoption-m
https://www.python.org/dev/peps/pep-0338/

"""
from nobubo import cli

if __name__ == "__main__":
    # https://github.com/pallets/click/issues/1399
    cli.main(prog_name="nobubo")



import logging
import pathlib
import re
from typing import List, Tuple, Optional

import pikepdf

from nobubo import errors
from nobubo.assembly import NobuboInput, PageSize, Layout
from nobubo.disassembly import NobuboOutput


logger = logging.getLogger(__name__)


def parse_cli_input_data(
    input_layout: List[Tuple[int, int, int]],
    reverse_assembly: bool,
    input_path: str,
) -> NobuboInput:
    try:
        with pikepdf.open(pathlib.Path(input_path)) as inputfile:
            logger.info(
                f"Received a pdf with {len(input_layout)} overview(s) "
                f"and {len(inputfile.pages)} pages."
            )
            # first page (getPage(0)) may contain overview, so get second one
            width, height = page_dimensions(inputfile.pages[1])

            input_properties = NobuboInput(
                input_filepath=pathlib.Path(input_path),
                number_of_pages=len(inputfile.pages),
                pagesize=PageSize(width=width, height=height),
                layout=parse_input_layouts(input_layout),
                reverse_assembly=reverse_assembly,
            )
            logger.debug(f"Parsed input properties: {input_properties}")
    except OSError as e:
        raise errors.UsageError(
            f"While reading the input pdf file, " f"this error occurred:\n{e}"
        )
    return input_properties


def parse_input_layouts(input_layout: List[Tuple[int, int, int]]) -> List[Layout]:
    return [
        Layout(first_page=data[0], columns=data[1], rows=data[2])
        for data in input_layout
    ]


def parse_cli_output_data(
    output_layout_cli: str,
    print_margin: int,
    output_path: str,
) -> NobuboOutput:
    output_properties = NobuboOutput(
        output_path=pathlib.Path(output_path),
        output_pagesize=parse_output_layout(output_layout_cli, print_margin)
        if output_layout_cli
        else None,
    )
    logger.debug(f"Parsed output properties: {output_properties}")
    return output_properties


def parse_output_layout(
    output_layout_cli: str, print_margin: Optional[int] = None
) -> PageSize:
    print_size: List[int] = []
    if output_layout_cli == "a0":
        print_size = to_mm("841x1189")
    if output_layout_cli == "us":  # Arch E /Arch 6 size of 36 × 48 inches
        print_size = to_mm("914x1220")
    elif "x" in output_layout_cli:
        print_size = to_mm(output_layout_cli)

    if print_margin:
        return to_userspaceunits([size - (2 * print_margin) for size in print_size])
    else:
        return to_userspaceunits(print_size)


def page_dimensions(page: pikepdf.Page) -> Tuple[float, float]:
    """
    Calculate the x, y value for the offset in default user space units
    as defined in the pdf standard.
    :param page: A PDF page.
    :return: tuple with x, y value.
    """
    if not hasattr(page, "cropbox"):
        # page is of type Object, and either mediabox, cropbox or trimbox
        # are all of type pikepdf.objects.Object
        # they exist (or not) depending on the pdf itself
        box: pikepdf.objects.Array = page.mediabox
    else:
        box = page.cropbox
    # pikepdf has some confusing types here
    # type(box[2]) returns decimal.Decimal, but mypy/pikepdf insists it's of type Object
    # the workaround is to convert it to a string,
    # then to a float to calculate the offset
    # since both Decimal and Object allow the stringification
    width = float(str(box[2])) - float(str(box[0]))
    height = float(str(box[3])) - float(str(box[1]))
    return round(width, 2), round(height, 2)


def to_mm(output_layout: str) -> List[int]:
    ol_in_mm = re.compile(r"\d+[x]\d+").findall(output_layout)[0].split("x")
    return [int(x) for x in ol_in_mm]


def to_userspaceunits(width_height: List[int]) -> PageSize:
    """
    Convert a page's physical width and height from millimeters to
    default user space unit, which is defined in the pdf standard as 1/72 inch.

    :param width_height: Width and height of the physical page in millimeters (mm),
    on which the pattern will be printed.
    :return: Width and height of the physical page in default user space units.
    """
    # 1 mm = 5/127 inches = 0.03937 inches;  1/72 inch = 0.013888889
    # conversion factor = 5/127 / 1/72 = 360/127 = 2.834645669
    conversion_factor = 2.834645669

    return PageSize(
        width=(round(width_height[0] * conversion_factor, 3)),
        height=(round(width_height[1] * conversion_factor, 3)),
    )
